/*
 * File:        main.c
 * Author:      Matthew Magaldi and Aidan Angus
 * Adapted from:
 *              main.c by
 * Author:      Syed Tahmid Mahbub
 * Target PIC:  PIC32MX250F128B
 */

/*
 * //////////////////////////////////
 PIN CONNECTIONS
 * 
 microSD - SPI1
 SCK - SCK1, PIN 25
 CS - RPB9, PIN 18
 DO - RPB11, PIN 22
 DI - RPB8, PIN 17
 CD - RPB10, PIN 21
 
 accelerometer
 X - AN4, PIN 6
 Y - AN11, PIN 24
 Z - AN5, PIN 7
 
 pressure - SPI2
 SCK - SCK2 (2MHZ)
 CS - RPB7, PIN 16
 DI - RPA2, PIN 9
 DO - RPA4, PIN 12
 
 gyro - SPI2, share spi but with different CS
 SCK - SCK2 (2MHZ)
 CS - RPB5, PIN 14
 DI - RPA2, PIN 9
 DO - RPA4, PIN 12
 */

////////////////////////////////////
// clock AND protoThreads configure!
// You MUST check this file!
#include "config.h"
// threading library
#include "pt_cornell_1_2_1.h"
#include <time.h>

////////////////////////////////////
// spi pressure
#include "pressure.h"
// need for rand function
#include <stdlib.h>
#include <math.h>

//SD Card includes
#include "FSIO.h"
#include "SD-SPI.h"
#include "sd_reader.h"

//Gyro includes
#include "gyro.h"
////////////////////////////////////

// define setup parameters for OpenADC10
// Turn module on | ouput in integer | trigger mode auto | enable autosample
// ADC_CLK_AUTO -- Internal counter ends sampling and starts conversion (Auto convert)
// ADC_AUTO_SAMPLING_ON -- Sampling begins immediately after last conversion completes; SAMP bit is automatically set
// ADC_AUTO_SAMPLING_OFF -- Sampling begins with AcquireADC10();
#define PARAM1  ADC_FORMAT_INTG16 | ADC_CLK_AUTO | ADC_AUTO_SAMPLING_ON //

// define setup parameters for OpenADC10
// ADC ref external  | disable offset test | disable scan mode | do 1 sample | use single buf | alternate mode off
#define PARAM2  ADC_VREF_AVDD_AVSS | ADC_OFFSET_CAL_DISABLE | ADC_SCAN_OFF | ADC_SAMPLES_PER_INT_1 | ADC_ALT_BUF_OFF | ADC_ALT_INPUT_OFF

// Define setup parameters for OpenADC10
// use peripherial bus clock | set sample time | set ADC clock divider
// ADC_CONV_CLK_Tcy2 means divide CLK_PB by 2 (max speed)
// ADC_SAMPLE_TIME_5 seems to work with a source resistance < 1kohm
#define PARAM3 ADC_CONV_CLK_PB | ADC_SAMPLE_TIME_5 | ADC_CONV_CLK_Tcy2 //ADC_SAMPLE_TIME_15| ADC_CONV_CLK_Tcy2

// define setup parameters for OpenADC10
// set AN4 and  as analog inputs
#define PARAM4	ENABLE_AN4_ANA |  ENABLE_AN11_ANA | ENABLE_AN5_ANA

// define setup parameters for OpenADC10
// do not assign channels to scan
#define PARAM5	SKIP_SCAN_ALL


void calibrate(void);
void init_spi2(void);

// string buffer
char buffer[60];

// === thread structures ============================================
// thread control structs
// note that UART input and output are threads
static struct pt pt_accel, pt_pressure, pt_sd, pt_gyro;

// === the fixed point macros ========================================
typedef signed int fix16 ;
#define multfix16(a,b) ((fix16)(((( signed long long)(a))*(( signed long long)(b)))>>16)) //multiply two fixed 16:16
#define float2fix16(a) ((fix16)((a)*65536.0)) // 2^16
#define fix2float16(a) ((float)(a)/65536.0)
#define fix2int16(a)    ((int)((a)>>16))
#define int2fix16(a)    ((fix16)((a)<<16))
#define divfix16(a,b) ((fix16)((((signed long long)(a)<<16)/(b)))) 
#define sqrtfix16(a) (float2fix16(sqrt(fix2float16(a)))) 
#define absfix16(a) abs(a)
#define truncfix16(a) (a & 0xFF00) //bit-wise AND operator for fix 16

static volatile int time_microsec = 0;
// == Timer 2 ISR =====================================================
// just toggles a pin for timeing strobe
void __ISR(_TIMER_2_VECTOR, ipl2) Timer2Handler(void)
{
    //overflow in 30 minutes
    time_microsec ++;
    mT2ClearIntFlag();
}

// === Accelerometer Thread =================================================
static float ax, ay, az;
static float vx = 0, vy = 0, vz = 0;
static float dt = 0.013;
// 0g calibrations
static volatile float x0= 0, y0 = 0, z0 = 0;
//-3g is at 0v and 3g is at 3.3V
//-3g is at 0 and 3g is at 1024 with full scaling in-between
static int time_begin, cycle;
static PT_THREAD (protothread_accel(struct pt *pt))
{
    PT_BEGIN(pt);
    while(1) {
        time_begin = time_microsec;
        // use ground as neg ref for A | use AN4 for input A     
        // configure to sample AN9 
        SetChanADC10( ADC_CH0_NEG_SAMPLEA_NVREF | ADC_CH0_POS_SAMPLEA_AN4 ); // configure to sample AN4 
        PT_YIELD_TIME_msec(2);
        ax = ReadADC10(0);   // read the result of channel 9 conversion from the idle buffer
        
        SetChanADC10( ADC_CH0_NEG_SAMPLEA_NVREF | ADC_CH0_POS_SAMPLEA_AN5 ); // configure to sample AN4 
        PT_YIELD_TIME_msec(2);
        ay = ReadADC10(0);   // read the result of channel 9 conversion from the idle buffer

        SetChanADC10( ADC_CH0_NEG_SAMPLEA_NVREF | ADC_CH0_POS_SAMPLEA_AN11 ); // configure to sample AN4 
        PT_YIELD_TIME_msec(2);
        az = ReadADC10(0);   // read the result of channel 9 conversion from the idle buffer
        
        cycle = time_microsec - time_begin;
        //velocity calculation
        vx += (ax-x0)/94*9.8*cycle/100000;
        vy += (ay-y0)/94*9.8*cycle/100000;
        
      }
    PT_END(pt);
} // timer thread

// === PRESSURE THREAD ==============================================
// retrieve information
static volatile float t=12, p=1000, a=200;
static PT_THREAD (protothread_pressure(struct pt *pt))
{
    PT_BEGIN(pt);
    while(1) {
        PT_YIELD_TIME_msec(1000);
        t = get_temperature();
        p = get_pressure();
        a = get_altitude();
    }
    PT_END(pt);
}

// === GYRO THREAD ==============================================
// retrieve information
static volatile float gx,gy,gz;
static PT_THREAD (protothread_gyro(struct pt *pt))
{
    PT_BEGIN(pt);
    while(1) {
        PT_YIELD_TIME_msec(50);
        gx = (float)getGyro(OUT_X);
        gy = (float)getGyro(OUT_Y);
        gz = (float)getGyro(OUT_Z);
    }
    PT_END(pt);
}

// === SD Storage THREAD ==============================================
// send information to the SD card for storage
FSFILE * myfile;
char str[120];
static PT_THREAD (protothread_sd(struct pt *pt))
{
    PT_BEGIN(pt);
    while(1) {
        PT_YIELD_TIME_msec(100);
        //Data order: ax, ay, az, gx, gy, gz, temp, press, alt, time
        sprintf(str, "%f, %f, %f, %f, %f, %f, %f, %f, %f, %u \r\n", ax, ay, az, gx, gy, gz, t, p, a, time_tick_millsec);
        myfile = FSfopen("dataCar.csv", FS_APPENDPLUS);
        FSfwrite(str, 1, 120, myfile);
        FSfclose(myfile);
    }
    PT_END(pt);
} // sd thread

// === Main  ======================================================
void main(void) {
    //SYSTEMConfigPerformance(PBCLK);
  
    ANSELA = 0; ANSELB = 0;
    
    OpenTimer2(T2_ON | T2_SOURCE_INT | T2_PS_1_1, 400);
    ConfigIntTimer2(T2_INT_ON | T2_INT_PRIOR_2);
    mT2ClearIntFlag(); // and clear the interrupt flag
    
    //mPORTBSetPinsAnalogIn(BIT_14 | BIT_13 | BIT_3);
	CloseADC10();	// ensure the ADC is off before setting the configuration
    // use ground as neg ref for A | use AN4 for input A     
    // configure to sample AN9 
    SetChanADC10( ADC_CH0_NEG_SAMPLEA_NVREF | ADC_CH0_POS_SAMPLEA_AN11 ); // configure to sample AN4 
    OpenADC10( PARAM1, PARAM2, PARAM3, PARAM4, PARAM5 ); // configure ADC using the parameters
    EnableADC10(); // Enable the ADC
    
    // === config threads ==========
    // turns OFF UART support and debugger pin, unless defines are set
    PT_setup();

    // === setup system wide interrupts  ========
    INTEnableSystemMultiVectoredInt();
    
    // Initialize the threads
    PT_INIT(&pt_accel);
    PT_INIT(&pt_pressure);
    PT_INIT(&pt_sd);
    PT_INIT(&pt_gyro);
    
    //////////////////////////////////////////////////////////////
    //Initialize the Sensor Array
    //////////////////////////////////////////////////////////////
    mPORTBSetPinsDigitalOut(BIT_0);
    mPORTBClearBits(BIT_0);
    //Calibrate the Accelerometer
    calibrate();
    
    //Initialize the SPI-2 Channel
    init_spi2();
    
    //Initialize the temperature, pressure, and altitude sensor
    pressure_begin(BMP183_MODE_ULTRAHIGHRES); //specify the mode for the sensor

    //Initialize the gyro sensor
    while(!begin_gyro(L3DS20_RANGE_500DPS));   //make sure we initialize and gyro is responding
    
    //Initialize the SD Card 
    while (!FSInit()); //if card detected and  formatted correctly, 1 is returned
    mPORTBSetBits(BIT_0);
    
    // round-robin scheduler for threads
    while (1){
        PT_SCHEDULE(protothread_accel(&pt_accel));
        PT_SCHEDULE(protothread_pressure(&pt_pressure));
        PT_SCHEDULE(protothread_gyro(&pt_gyro));
        PT_SCHEDULE(protothread_sd(&pt_sd));
    }
} // main

/*
 Set the SPI1 DO, DI pins and open the Channel
 Used for the Pressure & Gyro Sensor
 */
void init_spi2(void) {
    //use channel 2, should be channel 2 now to debug with tft - leave out sd card
    //configure sdi and sdo pins
    RPA4R = 4; //RPA4      // SDO/MOSI pin for the SPI1DataOut - data from processor -> sensor
    SDI2R = 0; // RPA2       // SDI/MISO pin for the SPI1DataIn - data from sensor to processor
    
    //Set Chip Selects as outputs and high
    GYRO_CS_dir = 0;
    GYRO_CS = 1;
    PR_CS_dir = 0;
    PR_CS = 1;
    
    //Open the SPI Channel
    SpiChnOpen(2, SPI_OPEN_MSTEN | SPI_OPEN_MODE8 | SPI_OPEN_CKE_REV | SPI_OPEN_ON , PBCLK/SPI2_freq);
}


/*
 Calibrate the accelerometer when holding still
 * to zero on average value over 100 cycles
 */
void calibrate(void) {
    int cycles = 100;
    int i;
    int until;
    for (i = 0; i < cycles; i ++) {
        // use ground as neg ref for A | use AN4 for input A     
        // configure to sample AN9 
        SetChanADC10( ADC_CH0_NEG_SAMPLEA_NVREF | ADC_CH0_POS_SAMPLEA_AN4 ); // configure to sample AN4 
        until = time_tick_millsec + 2;
        while(time_tick_millsec < until);
        x0 += ReadADC10(0);   // read the result of channel 9 conversion from the idle buffer

        SetChanADC10( ADC_CH0_NEG_SAMPLEA_NVREF | ADC_CH0_POS_SAMPLEA_AN5 ); // configure to sample AN4 
        until = time_tick_millsec + 2;
        while(time_tick_millsec < until);
        y0 += ReadADC10(0);   // read the result of channel 9 conversion from the idle buffer

        SetChanADC10( ADC_CH0_NEG_SAMPLEA_NVREF | ADC_CH0_POS_SAMPLEA_AN11 ); // configure to sample AN4 
        until = time_tick_millsec + 2;
        while(time_tick_millsec < until);
        z0 += ReadADC10(0);   // read the result of channel 9 conversion from the idle buffer
    }
    x0 = x0/cycles;
    y0 = y0/cycles;
    z0 = z0/cycles;
}

// === end  ======================================================